from __future__ import annotations
import shutil
import os
import pathlib
import datetime
import subprocess
import platform
from typing import List, Any, Tuple, Callable

import dir_ops as do
import py_starter as ps


class Dir ( do.BaseDir ) :


    def __init__ ( self, *args, **kwargs ):

        do.BaseDir.__init__( self, *args, **kwargs )
        self.DIR_CLASS = Dir
        self.PATH_CLASS = Path
        self.DIRS_CLASS = Dirs
        self.PATHS_CLASS = Paths

    
    @staticmethod
    def open_dir( dir ) -> None:

        """Opens the dir in the file explorer """

        if platform.system() == 'Windows':
            os.startfile(dir)

        elif platform.system() == 'Darwin':
            subprocess.call(['open', dir])


    @staticmethod
    def exists_dir( dir: str, *args, **kwargs ) -> bool:
        assert os.path.exists( dir )

    @staticmethod
    def remove_dir( dir: str, *args, **kwargs ) -> None:
        shutil.rmtree(dir)

    @staticmethod
    def create_dir( dir: str, *args, **kwargs ) -> None:
        os.mkdir( dir )

    @staticmethod
    def copy_dir( dir: str, *args, destination: str = '', **kwargs) -> None:
        shutil.copytree( dir, destination )

    @staticmethod
    def list_files_dir( dir: str ) -> List[str]:

        contents = os.listdir( dir )

        for i in range(len(contents)-1,-1,-1):
            if not os.path.isfile( dir + '/' + contents[i] ):
                del contents[i]

        return contents

    @staticmethod
    def list_subfolders_dir( dir: str) -> List[str]:

        contents = os.listdir( dir )

        for i in range(len(contents)-1,-1,-1):
            if not os.path.isdir( dir + '/' + contents[i] ):
                del contents[i]

        return contents

    @staticmethod
    def get_size_dir( dir: str, *args, **kwargs ) -> float:

        self = Path( path = dir, **self.inherited_kwargs )
        Paths_inst = self.list_contents_Paths( block_dirs=True, block_paths=False )

        bytes = 0
        for Path_inst in Paths_inst:
            Path_inst.get_size( conversion = None )
            bytes += Path_inst.size 

        return bytes         


class Path( Dir, do.BasePath ):

    """
    Inherits from the Dir class found in Dir.py
    Path (uppercase P) is a Path class, path (lowercase P) is a string with file absolute path
    """

    def __init__( self, *args, **kwargs ):

        do.Dir.__init__( self, *args, **kwargs )
        do.BasePath.__init__( self, *args, **kwargs )        
        self.DIR_CLASS = Dir
        self.PATH_CLASS = Path
        self.DIRS_CLASS = Dirs
        self.PATHS_CLASS = Paths

    @staticmethod
    def exists_path( path: str, *args, **kwargs ) -> bool:
        assert os.path.exists( path )

    @staticmethod
    def copy_path( path: str, *args, destination: str = '', **kwargs ) -> None:
        shutil.copyfile(path, destination)

    @staticmethod
    def create_path( path: str, *args, **kwargs) -> None:

        self = Path( path )        
        if not self.write( string = '' ):
            assert False

    @staticmethod
    def remove_path(path: str, *args, **kwargs) -> None:
        os.remove(path)

    @staticmethod
    def rename_path( path, *args, destination: str = '', **kwargs) -> None:
        os.rename( path, destination )

    @staticmethod
    def get_size_path( path, *args, **kwargs ) -> float:
        return os.stat( path ).st_size

    @staticmethod
    def get_mtime_path( path, *args, **kwargs ) -> datetime.datetime:

        """get the time of modification as a datetime object"""

        mtime = pathlib.Path(path).stat().st_mtime
        dt = datetime.datetime.fromtimestamp( mtime )
        return dt

    @staticmethod
    def read_path( path, **kwargs ) -> Any:
        """reads from a text file at path, read py_starter.read_text_file() for kwargs """

        return ps.read_text_file( path, **kwargs )

    @staticmethod
    def write_path( path, **kwargs ) -> None:
        """writes to a text file at path, read py_starter.write_text_file() for kwargs """

        ps.write_text_file( path, **kwargs )

    @do.inherited_instance_method
    def import_module( self, *args, **kwargs) -> Callable :
        pass
    
    @staticmethod
    def import_module_path( path, *args, **kwargs ) -> Callable:
        return ps.import_module_from_path( path, **kwargs )


class Dirs( do.BaseDirs ):


    def __init__ ( self, *args, **kwargs ):

        do.BaseDirs.__init__( self, *args, **kwargs )
        self.DIR_CLASS = Dir
        self.PATH_CLASS = Path
        self.DIRS_CLASS = Dirs
        self.PATHS_CLASS = Paths


class Paths( Dirs, do.BasePaths ):

    def __init__( self, *args, **kwargs ):

        Dirs.__init__( self )
        do.BasePaths.__init__( self, *args, **kwargs )
        self.DIR_CLASS = Dir
        self.PATH_CLASS = Path
        self.DIRS_CLASS = Dirs
        self.PATHS_CLASS = Paths





import functools
import dir_ops as do
import py_starter as ps

def base_instance_method(method):

    """instance methods call the corresponding staticmethod 
    Example: Dir_instance.exists(*,**) calls Dir.exists_dir( Dir_instance.path,*,** )   
    """

    @functools.wraps(method)
    def wrapper( self, *called_args, **called_kwargs):

        new_method_name = method.__name__ + self.STATIC_METHOD_SUFFIX
        return self.get_attr( new_method_name )( self.path, *called_args, **called_kwargs )
        
    return wrapper

def inherited_instance_method(method):

    """instance methods call the corresponding staticmethod 
    Example: Dir_instance.exists(*,**) calls Dir.exists_dir( Dir_instance.path,*,** )   """

    @functools.wraps(method)
    def wrapper( self, *called_args, **called_kwargs):

        new_method_name = method.__name__ + self.STATIC_METHOD_SUFFIX
        instance_args = [ self.get_attr( attr ) for attr in self.INSTANCE_METHOD_ATTS ]

        return self.get_attr( new_method_name )( *instance_args, *called_args, **called_kwargs )
      
    return wrapper

def dirs_wrap( method_name, track_success = False ):

    def dirs_gen( method ):

        @functools.wraps( method )
        def wrapper( self, *args, **kwargs ):

            valid = True
            for DirPath in self:
                value = DirPath.get_attr( method_name )( *args, **kwargs )

                if value == False:
                    valid = False

            if track_success:
                return valid

        return wrapper

    return dirs_gen
  

###

def to_from_wrapper_factory( method, action_str, self, *args, override: bool = False, print_off: bool = False, 
                    Destination = None, destination = '', 
                    overwrite: bool = False,
                    **kwargs  ):

    if Destination == None:
        if self.type_path:
            Destination = do.Path( destination )
        if self.type_dir:
            Destination = do.Dir( destination )


    ######  Insert special instructions
    if action_str == 'download' or action_str == 'copy' or action_str == 'rename':
        Destination.create_parents()
    
    ######
    if not override:
        do.print_to_from( True, action_str, str(self), str(Destination) )
        override = ps.confirm_raw( string = '' )

    if override:
        ###### Check for conflicting destination locations
        if action_str == 'download' or action_str == 'copy' or action_str == 'rename':
            if self.exists() and Destination.exists():
                if not overwrite:
                    print ('ERROR: Destination ' +str(Destination)+ ' already exists. Pass "overwrite=True" to overwrite existing file.')
                    return False
                else:
                    Destination.remove()

        # perform the actual method        
        do.print_to_from( print_off, action_str, str(self), str(Destination) )

        if method( self, *args, destination = Destination.path, override=override, print_off=print_off, **kwargs ):
            return True
        else:
            do.print_to_from( True, action_str, str(self), str(Destination) )
            print ('ERROR: could not complete ' + action_str)

    return False



def download_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'download', self, *args, **kwargs )

    return wrapper

def upload_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'upload', self, *args, **kwargs )

    return wrapper

def copy_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'copy', self, *args, **kwargs )

    return wrapper

def rename_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'rename', self, *args, **kwargs )

    return wrapper


def remove_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'remove', self, *args, **kwargs )

    return wrapper


def create_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):
        return to_from_wrapper_factory( method, 'create', self, *args, **kwargs )

    return wrapper


###

def get_size_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, conversion=None, **kwargs ):

        size_bytes = method( self, *args, conversion=None, **kwargs )
        
        size, size_units = do.convert_bytes( size_bytes, conversion=conversion )
        self.size = size
        self.size_units = size_units

        return size, size_units

    return wrapper

def get_mtime_wrap( method ):

    @functools.wraps( method )
    def wrapper( self, *args, **kwargs ):

        mtime = method( self, *args, **kwargs )
        self.mtime = mtime

        return mtime

    return wrapper




DEBUG = False

from .support_functions import *
from .decorators import *
from .base_dir_ops import *
from .local_dir_ops import *
from .remote_dir_ops import *
import os

_Dir = Dir( os.path.abspath( __file__ ) ).ascend()   #Dir that contains the package 
_src_Dir = _Dir.ascend()                                  #src Dir that is one above
_repo_Dir = _src_Dir.ascend()                    
_cwd_Dir = Dir( get_cwd() )




from __future__ import annotations

import dir_ops as do
from parent_class import ParentClass
from parent_class import ParentPluralList

import datetime
from typing import List, Any, Tuple
import os
import py_starter as ps

class BaseDir( ParentClass ):

    STATIC_METHOD_SUFFIX = '_dir'
    INSTANCE_METHOD_ATTS = ['path']
    inherited_kwargs = {}

    def __init__( self, *args, **kwargs ):

        ParentClass.__init__( self )
        self.dir_construct( *args, **kwargs )

        self.DIR_CLASS = BaseDir
        self.PATH_CLASS = BasePath
        self.DIRS_CLASS = BaseDirs
        self.PATHS_CLASS = BasePaths

    def dir_construct( self, path = '', **kwargs ):

        self.path = do.replace_delims( path )
        self.path = do.remove_hanging_slashes( self.path )  # 'C:/Users/e150445/Documents/MO-EE/Data/Raw'
        self.dirs = do.path_to_dirs( self.path )                #[ 'C','Users','e150445','Documents','MO-EE','Data','Raw' ]
        self.type_dir = True
        self.type_path = False
        self.type_dirs = False
        self.type_paths = False

        self.size = None                          # don't init with checking the size, takes too long
        self.size_units = None

        #alias just for quick coding
        self.p = self.path

    def print_imp_atts( self, **kwargs ):

        return self._print_imp_atts_helper( atts = ['path','dirs'], **kwargs )

    def print_one_line_atts(self, **kwargs ):

        return self._print_one_line_atts_helper( atts = ['type','path'], **kwargs )

    def __eq__( self, other_Dir ):

        if isinstance( other_Dir, self.DIR_CLASS ):
            return self.path == other_Dir.path
        return False

    @staticmethod
    def is_Dir( Object: Any ) -> bool:

        """returns boolean if Object is a Dir"""
        return isinstance( Object, BaseDir )

    def ascend( self, level_to_ascend: int = 1 ) -> BaseDir:
        
        """go up a x number directories -> "levels_to_ascend" """

        return self.DIR_CLASS( path = do.join( *self.dirs[:-1*level_to_ascend] ), **self.inherited_kwargs )
    
    @staticmethod
    def ascend_dir( dir: str, levels_to_ascend: int = 1 ) -> str:

        """go up a x number directories -> "levels_to_ascend" """

        dirs = do.path_to_dirs( dir )
        return do.join( *dirs[:-1*levels_to_ascend] )

    @do.base_instance_method
    def join( self, *args, **kwargs ):
        pass
    
    @staticmethod
    def join_dir( dir: str, *other_dirs, **kwargs ):
        
        """add more dirs to the Dir path"""
        if dir != '':
            return do.join( dir, *other_dirs )
        else:
            return do.join( *other_dirs )

    def join_Dir( self, Dir: BaseDir = None, path: str = '' ) -> BaseDir:
        
        if Dir != None:
            path = Dir.path
        return self.DIR_CLASS( path = self.join( path ), **self.inherited_kwargs )

    def join_Path( self, Path: BasePath = None, path: str = '' ) -> BasePath:

        if Path != None:
            path = Path.path

        return self.PATH_CLASS( path = self.join( path ), **self.inherited_kwargs )

    ##################

    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def open( self, *args, **kwargs ):
        pass

    @staticmethod
    def open_dir( *args, **kwargs ): #should be defined by a child class
        assert False

    ##################

    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def exists( self, *args, **kwargs ) -> bool:
        pass
    
    @staticmethod
    def exists_dir( *args, **kwargs ): 
        return None

    ##################

    @do.remove_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def remove(self, *args, **kwargs):
        pass

    @staticmethod
    def remove_dir( *args, **kwargs ): 
        assert False

    ##################


    def create_parents(self, *args, **kwargs):

        """use recursion to travel all the way up the parent directories until we find one that exists, then unfold and create each directory"""

        parent_Dir = self.ascend()
        if parent_Dir.exists() or len(self.dirs) <= 1: #Parent Dir exists or we are at the base directory
            return

        else:
            parent_Dir.create_parents( *args, **kwargs )
            parent_Dir.create( *args, **kwargs )

    ##################

    @do.create_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def create(self, *args, **kwargs) -> bool:
        pass

    @staticmethod
    def create_dir( *args, **kwargs ) -> bool: #should be defined by a child class
        assert False

    ##################

    @do.copy_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def copy(self, *args, **kwargs) -> bool:
        pass
    
    @staticmethod
    def copy_dir( *args, **kwargs ) -> bool:
        assert False

    ##################

    def list_contents( self, *args, print_off: bool = False, **kwargs ):

        filenames = []
        filenames.extend( self.list_subfolders() )
        filenames.extend( self.list_files() )

        if print_off:
            ps.print_for_loop( filenames )
        
        return filenames

    ##################

    @do.inherited_instance_method
    def list_files( self, *args, **kwargs ):
        pass

    @staticmethod
    def list_files_dir( *args, **kwargs ):
        return []

    ##################

    @do.inherited_instance_method
    def list_subfolders( self, *args, **kwargs ):
        pass

    @staticmethod
    def list_subfolders_dir( *args, **kwargs ):
        return []

    ##################

    @do.get_size_wrap
    @do.inherited_instance_method
    def get_size( self, *args, **kwargs ):
        pass

    @staticmethod
    def get_size_dir( *args, **kwargs ):
        return 0

    ##################

    def get_rel( self, other_Dir: BaseDir ) -> BaseDir:

        """Given a Dir object, find the relative Dir from Dir to self"""

        return self.DIR_CLASS( path = self.get_rel_dir( self.path, other_Dir.path ), **self.inherited_kwargs )

    @staticmethod
    def get_rel_dir( dir, other_dir ) -> str:
        
        if dir != '':
            return os.path.relpath( dir, other_dir )
        else:
            return other_dir
    
    ##################

    def list_contents_Paths( self, block_dirs: bool = True, block_paths: bool = False, **kwargs ) -> BasePaths:

        Paths_inst = self.PATHS_CLASS( **self.inherited_kwargs )

        # 1. Add all files
        if not block_paths:
            paths = self.list_files( **kwargs )

            for path in paths:
                Paths_inst._add( self.join_Path( path = path ) )

        # 2. Add all dirs
        if not block_dirs:
            dirs = self.list_subfolders( **kwargs )

            for dir in dirs:
                Paths_inst._add( self.join_Dir( path = dir ) )
          
        return Paths_inst


    def walk( self, folders_to_skip: List[str] = ['.git'], **kwargs ) -> BasePaths:

        """Walk through all the contents of the directory"""

        Paths_inst = self.PATHS_CLASS( **self.inherited_kwargs )
        Paths_inst._add( self )

        Paths_under = self.list_contents_Paths( block_dirs = False, block_paths = False, **kwargs )

        for Path_inst in Paths_under:

            # all Paths are also Dirs
            if Path_inst.type_path:
                Paths_inst._add( Path_inst )

            elif Path_inst.type_dir:
                
                if Path_inst.dirs[-1] not in folders_to_skip:
                    Paths_inst.merge( Path_inst.walk( folders_to_skip = folders_to_skip, **kwargs ) )

        return Paths_inst

    def walk_contents_Paths( self, block_dirs: bool = True, block_paths: bool = False, folders_to_skip: List[str] = ['.git'], **kwargs ) -> BasePaths:

        """get all Paths and/or Dirs underneath the entire directory, optional params for returning paths and/or dirs"""

        Paths_inst = self.walk( folders_to_skip=folders_to_skip, **kwargs )
        keep_Paths = self.PATHS_CLASS( **self.inherited_kwargs )

        for Path_inst in Paths_inst:
            if Path_inst.type_path and not block_paths:
                keep_Paths._add( Path_inst )

            if Path_inst.type_dir and not block_dirs:
                keep_Paths._add( Path_inst )

        return keep_Paths

    def get_unique_Path( self, filename: str, **list_contents_Paths_kwargs ) -> str:

        """finds a unique Path for the proposed filename based on the contents of the Directory
        if file.txt already exists in the dir, return file1.txt or file2.txt, etc  """

        # 
        Paths_in_Dir = self.list_contents_Paths( block_dirs=True, block_paths=False, **list_contents_Paths_kwargs )
        filenames = [ P.filename for P in Paths_in_Dir ]

        filename_Path = self.join_Path( path = filename )
        if filename not in filenames:
            return filename_Path
        
        counter = 0
        while True:
            proposed_filename = filename_Path.root + str(counter) + filename_Path.extension
            if proposed_filename not in filenames:
                return self.join( path = proposed_filename )

            counter += 1


class BasePath( BaseDir ):

    STATIC_METHOD_SUFFIX = '_path'
    INSTANCE_METHOD_ATTS = ['path']

    def __init__( self, *args, **kwargs ):

        ParentClass.__init__( self )
        BaseDir.__init__( self, *args, **kwargs ) 
        self.path_construct()

        self.DIR_CLASS = BaseDir
        self.PATH_CLASS = BasePath
        self.DIRS_CLASS = BaseDirs
        self.PATHS_CLASS = BasePaths

    def path_construct(self):

        """Since the Path is different from a Dir, add the extra attiributes"""

        self.filename =     self.get_filename( self.path )                  # feb_mar_v1.0.txt
        self.root =         self.get_root( self.path )                      # feb_mar_v1
        self.root_dots =    self.get_root( self.path, allow_dots=True )     # feb_mar_v1.0
        self.ending =       self.get_ending( self.path )                    # txt
        self.extension =    '.' + self.ending                               # .txt
        self.size = None                          # don't init with checking the size, takes too long
        self.size_units = None
        self.mtime = None                        # a datetime object when its ready

        self.parent_Dir = self.ascend()
        self.type_dir = False
        self.type_path = True
        self.type_dirs = False
        self.type_paths = False

    def __eq__( self, other_Path: BasePath ) -> bool:

        """checks if self is equal to other_Path, returns bool"""

        if isinstance( other_Path, self.PATH_CLASS ):
            return self.path == other_Path.path
        return False

    @staticmethod
    def is_Path( Object: Any ) -> bool:

        """returns boolean if Object is a Dir"""
        return isinstance( Object, BasePath )

    def print_imp_atts(self, print_off = True):

        return self._print_imp_atts_helper( atts = ['path','dirs','ending','size'], print_off = print_off )

    @staticmethod
    def exists_path( *args, **kwargs ):
        return False
    
    @staticmethod
    def remove_path( *args, **kwargs ):
        assert False

    @do.rename_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def rename( self, *args, **kwargs):
        pass
    
    @staticmethod
    def rename_path( *args, **kwargs ):
        assert False

    @do.get_mtime_wrap
    @do.inherited_instance_method
    def get_mtime( self, *args, **kwargs ):
        pass
    
    @staticmethod
    def get_mtime_path( *args, **kwargs ) -> datetime.datetime:
        return datetime.datetime.utcfromtimestamp( 0 )

    @staticmethod
    def create_path( *args, **kwargs ) -> bool:
        assert False

    @do.inherited_instance_method
    def read( self, *args, **kwargs):
        pass
    
    @staticmethod
    def read_path( path, **kwargs ) -> Any:
        return None

    @ps.try_operation_wrap( debug = do.DEBUG )
    @do.inherited_instance_method
    def write( self, *args, **kwargs):
        pass

    @staticmethod
    def write_path( *args, **kwargs ) -> bool:
        assert False

    def smart_format( self, *args, formatting_dict, write = True, **kwargs) -> str:

        string = self.read()
        formatted_string = ps.smart_format( string, formatting_dict, **kwargs )

        if write:
            self.write( string = formatted_string )

        return formatted_string

    def get_rel( self, Dir_inst: BaseDir, path = '' ) -> BasePath:

        """Given a Dir object, find the relative Path from Dir to self"""

        if Dir_inst != None:
            path = Dir_inst.path

        return self.PATH_CLASS( path = self.get_rel_path( self.path, path ), **self.inherited_kwargs )

    @staticmethod
    def get_rel_path( path: str, dir: str ) -> str:
        return os.path.relpath( path, dir )

    @staticmethod
    def get_filename( path: str ) -> str:

        """returns the filename ('file.txt') from a long path ('C:/path/to/file.txt') """

        dirs = do.path_to_dirs(path)

        if len(dirs) > 0:
            return dirs[-1]
        else:
            return ''
    
    @staticmethod
    def get_root( path: str, allow_dots: bool = False ) -> str:

        '''returns the root of the filename from a path  Dir/a_file1.txt returns "a_file1" '''
        filename = BasePath.get_filename(path)

        if allow_dots:
            root = '.'.join( filename.split('.')[:-1] )

        else:
            root = filename.split('.')[0]
        
        return root
    
    @staticmethod
    def get_ending( path: str ) -> str:

        '''returns the file ending from a path'''
        
        filename = BasePath.get_filename(path)
        ending = filename.split('.')[-1]
        return ending



class BaseDirs( ParentPluralList ):

    inherited_kwargs = {}

    def __init__( self, *args, given_Dirs = [], given_Paths = [], **kwargs ):

        ParentPluralList.__init__( self, att = 'Objs' )
        self.DIR_CLASS = BaseDir
        self.PATH_CLASS = BasePath
        self.DIRS_CLASS = BaseDirs
        self.PATHS_CLASS = BasePaths

        self.Dirs = self.Objs # Make an Alias

        for D in given_Dirs:
            self._add( D )
        for D in given_Paths:
            self._add( D )
        
        self.type_dir = False
        self.type_path = False
        self.type_dirs = True
        self.type_paths = False

    @staticmethod
    def is_Dirs( Object: Any ) -> bool:

        """returns boolean if Object is a Dir"""
        return isinstance( Object, BaseDirs )

    def join_Dir( self, Dir_inst: BaseDir ) -> BasePaths:

        """Joins Dir_inst to each Dir/Path contained in the Object"""

        Paths_inst = self.PATHS_CLASS( **self.inherited_kwargs )

        for DirPath in self:

            if DirPath.type_path:
                Paths_inst._add( Dir_inst.join_Path( DirPath ) )
            
            if DirPath.type_dir:
                Paths_inst._add( Dir_inst.join_Dir( DirPath ) )

        return Paths_inst

    def merge( self, other_Dirs: Any ) -> None:

        """add all Dir objects from another Dirs instance to self"""

        for Dir_inst in other_Dirs:
            self._add( Dir_inst )

    def export_strings( self ) -> List[str]:

        """Returns all the paths of each Dir/Path contained"""

        return [ O.path for O in self ]

    @do.dirs_wrap( 'create', track_success=True )
    def create( self, *args, **kwargs ):
        pass

    @do.dirs_wrap( 'remove', track_success=True )
    def remove( self, *args, **kwargs ):
        pass


class BasePaths( BaseDirs ):

    def __init__ ( self, *args, **kwargs ):

        BaseDirs.__init__( self, **kwargs )
        self.Paths = self.Objs # Make an Alias

        self.DIR_CLASS = BaseDir
        self.PATH_CLASS = BasePath
        self.DIRS_CLASS = BaseDirs
        self.PATHS_CLASS = BasePaths

        self.type_dir = False
        self.type_path = False
        self.type_dirs = False
        self.type_paths = True

    @staticmethod
    def is_Paths( Object: Any ) -> bool:

        """returns boolean if Object is a Dir"""
        return isinstance( Object, BasePaths )

    def get_rels( self, Dir_inst ) -> BasePaths:

        """Given a Dir object, find the relative Paths from Dir to the Paths"""

        Paths_inst = self.PATHS_CLASS( **self.inherited_kwargs )
        for P in self:
            Paths_inst._add( P.get_rel( Dir_inst ) )

        return Paths_inst

   


from typing import List, Any, Tuple
import os
import platform

DELIM = '/'
SECONDARY_DELIM = '\\'

def get_cwd() -> str:

    """returns current working directory"""

    cd = os.getcwd()

    #replace the cd with primary delim
    cd_replaced = join( * path_to_dirs(cd) )
    return cd_replaced

def path_to_dirs( path: str ) -> List[str]:

    """splits a path into directories"""
    if path == '':
        return []
    return path.split( DELIM )

def is_file( path: str ) -> bool :

    """returns a boolean value to check if the path is a file"""
    return os.path.isfile( path )

def is_dir( path: str ) -> bool :

    """returns a boolean value to check if the path is a directory"""
    return os.path.isdir( path )

def replace_delims( path: str, secondary_delim: str = SECONDARY_DELIM, delim: str = DELIM ) -> str :

    """turns data\dir\file.txt into data/dir/file.txt"""
    return path.replace( secondary_delim, delim )

def join( *items: str ) -> str:

    """joins a list of dirs into a path"""
    return DELIM.join( items )

def convert_bytes( bytes: int, conversion: str = 'MB' ) -> Tuple[ Any, Any ]:

    try:
        conversion = conversion.upper()
    except:
        pass

    kb_to_b = 1024

    conversion_factors = ['KB','MB','GB','TB','PB','EB']
    if conversion in conversion_factors:

        power = conversion_factors.index(conversion) + 1
        return bytes / (kb_to_b ** power), conversion

    return bytes, None


def add_prefix_to_paths( prefix_path, relative_paths ):

    """adds prefix to a list of relaive paths"""

    full_paths = []
    for relative_path in relative_paths:
        full_paths.append( join(prefix_path, relative_path) )

    return full_paths

def remove_prefix_from_paths( prefix_path, full_paths ):

    """Removes prefix from full paths"""

    number_of_dirs = len( path_to_dirs(prefix_path) )

    relative_paths = []
    for full_path in full_paths:

        folder_list = path_to_dirs( full_path )
        relative_list = folder_list[ number_of_dirs : ]

        relative_paths.append( join(*relative_list) )

    return relative_paths

def remove_hanging_slashes( path ):

    """removes ending slashes from paths
    turns "asdf/asdf//" into "asdf/sadf"   """

    # dont go through all the
    for i in range( len(path)-1, -1, -1 ):

        if path[i] != '/':
            return path[:i+1]

    return path

def get_desktop_dir() -> Any:

    """get the location of the desktop"""

    if platform.system() == 'Windows':
        import winshell
        return winshell.desktop()

    return None

def create_shortcut( target_Path, shortcut_Dir ) -> None:

    """places a shortcut from target_Path to shortcut_Dir"""

    if platform.system() == 'Windows':
        import winshell

        shortcut_path = shortcut_Dir.join( target_Path.root + '.lnk' )
        winshell.CreateShortcut( Path=shortcut_path, Target=target_Path.p )

    else:
        print ('No instructions for current OS')

def get_env_var_path_delim() -> str:

    """Returns the delimitter for the OS's path environment variables"""

    if platform.system() == 'Windows':
        return ';'
    elif platform.system() == 'Linux':
        return ':'

def split_env_var_paths( string: str ) -> List[str]:

    """returns a list of path strings split on the system delimitter"""

    return string.split( get_env_var_path_delim() )

def join_env_var_paths( paths: List[str] ) -> str:

    """ returns 'C:/Path1;C:/Path2' """
    return get_env_var_path_delim().join( paths )

def print_to_from( print_off: bool, action_str: str, from_str: str, to_str: str ):

    """ Copying C:/Users/path/file.txt ->  C:/Users/newfile.txt """

    if print_off:
        print ( "{action_str} \t {from_str} -> \t {to_str}".format( action_str = action_str, to_str = to_str, from_str = from_str ) )



import dir_ops as do
import py_starter as ps

class RemoteDir ( do.BaseDir ) :

    def __init__ ( self, *args, **kwargs ):

        do.BaseDir.__init__( self, *args, **kwargs )
        self.DIR_CLASS = RemoteDir
        self.PATH_CLASS = RemotePath
        self.DIRS_CLASS = RemoteDirs
        self.PATHS_CLASS = RemotePaths

    @do.upload_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )   
    @do.inherited_instance_method
    def upload( self, *args, **kwargs ):
        pass

    @staticmethod
    def upload_dir( *args, **kwargs ):
        assert False

    @do.download_wrap
    @ps.try_operation_wrap( debug = do.DEBUG )   
    @do.inherited_instance_method
    def download( self, *args, **kwargs ):
        pass
    
    @staticmethod
    def download_dir( *args, **kwargs ):
        assert False
   


class RemotePath( RemoteDir, do.BasePath ):

    def __init__( self, *args, **kwargs ):

        RemoteDir.__init__( self, *args, **kwargs )
        do.BasePath.__init__( self, *args, **kwargs )
        self.DIR_CLASS = RemoteDir
        self.PATH_CLASS =  RemotePath
        self.DIRS_CLASS = RemoteDirs
        self.PATHS_CLASS = RemotePaths

    @staticmethod
    def upload_path( *args, **kwargs ):
        assert False

    @staticmethod
    def download_path( *args, **kwargs ):
        assert False


class RemoteDirs( do.BaseDirs ):

    def __init__( self, *args, **kwargs ):

        do.BaseDirs.__init__( self, *args, **kwargs )
        self.DIR_CLASS = RemoteDir
        self.PATH_CLASS =  RemotePath
        self.DIRS_CLASS = RemoteDirs
        self.PATHS_CLASS = RemotePaths




class RemotePaths( RemoteDirs, do.BasePaths ):

    def __init__( self, *args, **kwargs ):

        RemoteDirs.__init__( self )
        do.Paths.__init__( self, *args, **kwargs )
        self.DIR_CLASS = RemoteDir
        self.PATH_CLASS =  RemotePath
        self.DIRS_CLASS = RemoteDirs
        self.PATHS_CLASS = RemotePaths

